#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "desc.h"
#include "usb.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )



#define TYPE_CONT	0
#define TYPE_BULK	2
#define TYPE_INT	3

#define SIZE_8		0
#define SIZE_32		2

#define ONE_BANK	0
#define DUAL_BANK	1


#define EP_CTRL		0
#define EP_CTRL_SZ_SFT	SIZE_8
#define EP_CTRL_SIZE	(8<<EP_CTRL_SZ_SFT)


int8_t usb_connected;
uint8_t reply_requested;

void send_zlp () {
	bit_clear(UEINTX, TXINI);  // send ZLP 
	while (bit_is_clear(UEINTX, TXINI))
		;
}

//#define USE_DUAL_BANK_BULK	0

int endpoint_setup(uint8_t ep) {
	char *p;
	uint8_t dir,attr,size,limit;
	uint8_t sz,bank;
	p = usbDescrConfig;
	limit = __LPM(p+2);
	while (p - usbDescrConfig < limit) {
		if ((__LPM(p+1) == USBDESCR_ENDPOINT) 
		     && ((__LPM(p+2) & 0xf) == ep))  { // endpoint desc.
			dir = (__LPM(p+2) & 0x80) == 0x80; // 0x80 : IN
			attr = __LPM(p+3); // bulk:2 , interrupt:3
			size = __LPM(p+4);
			sz = SIZE_8;
			while ((8<<sz) < size )
				sz++;
			bank = ONE_BANK;
#if (USE_DUAL_BANK_BULK)
			if ((ep >= 3) && (attr == 2)) {
				bank = DUAL_BANK;
			}
#endif
			UENUM = ep;
			if (bit_is_clear(UESTA0X, CFGOK)) {
				bit_set(UECONX, EPEN);
				UECFG0X = (attr<<EPTYPE0) | dir;
				UECFG1X = (sz<<EPSIZE0) | (bank<<EPBK0);
				bit_set(UECFG1X,ALLOC);
				{
					uint8_t buf[4];
					buf[0] = UECFG0X;
					buf[1] = UECFG1X;
					buf[2] = UESTA0X;
					buf[3] = UESTA1X;
					eeprom_write_block(buf,
					    (uint8_t *)(0x80 + (ep << 2)),4);
				}
			} else {
				bit_set(UERST,ep);
				nop();
				bit_clear(UERST,ep);
				bit_set(UECONX, EPEN);
			}
			return size;
		}
		p += __LPM(p);
	}
	return 0;
}

void start_pll() {
	if (bit_is_clear(PLLCSR,PLOCK)) {
#if F_CPU==8000000UL
		PLLCSR = (0<<PLLP0) | (1<<PLLE);
#else
		PLLCSR = (1<<PLLP0) | (1<<PLLE);
#endif
		while (bit_is_clear(PLLCSR,PLOCK))
			;
	}
	bit_clear(USBCON,FRZCLK); // enable usb clock 
	while (bit_is_clear(UDINT, SOFI))
		;
}

void usbInit() {
	int8_t i;
	USBCON = (1<<FRZCLK) | (1<<DETACH);
	UDIEN = 0;
	UEIENX = 0;
	bit_set(USBCON,USBE);     // enable usb
	bit_clear(UDCON, DETACH); // connect internal pull-up 
	start_pll();
#if 1
	bit_set(UDCON, RSTCPU);   // reset cpu @ USB_RESET
#else
	bit_clear(UDCON, RSTCPU);   // reset cpu @ USB_RESET
#endif
	//bit_set(UDIEN, SUSPE);    // Enable Suspend int
	bit_set(UDIEN, EORSTE);	  // Enable Reset int
	//bit_set(UDIEN, WAKEUPE);  // Enable Wakeup int (one time)
	UENUM = EP_CTRL;
	if (bit_is_clear(UESTA0X, CFGOK)) {
		bit_set(UECONX, EPEN);
		UECFG0X = (TYPE_CONT<<EPTYPE0) | (0<<EPDIR);
		UECFG1X = (EP_CTRL_SZ_SFT<<EPSIZE0) | (ONE_BANK<<EPBK0);
		bit_set(UECFG1X,ALLOC);
	} else {
		bit_set(UERST,EP_CTRL);
		nop();
		bit_clear(UERST,EP_CTRL);
		bit_set(UECONX, EPEN);
	}
	for (i=1; i<= 4; i++) {
		endpoint_setup(i);
	}
}

SIGNAL(USB_GEN_vect) {
	if (bit_is_set(UDINT,EORSTI)) {
		// USB Reset
		//usbInit();
		bit_clear(UDINT,EORSTI);
	}
	if (bit_is_set(UDINT, WAKEUPI)) {
		bit_clear(UDIEN, WAKEUPE);
		start_pll();
		bit_clear(UDINT, WAKEUPI);
	}
}

void usb_reply_setup(uint8_t *p, uint8_t len, int8_t mode) {
	uint8_t fifo_len = 0;
	uint8_t written = 0;
	uint8_t requested = reply_requested;

	bit_clear(UEINTX, NAKOUTI);
	while (len) {
		if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
			goto ret;
		}
		UEDATX = (mode == 0)?(*p):(__LPM(p));
		p++;
		fifo_len++;
		written++;
		len--;
		if (written >= reply_requested) break;
		if (fifo_len == EP_CTRL_SIZE) {
			bit_clear(UEINTX, TXINI); // send
			while (bit_is_clear(UEINTX, TXINI)
					&& bit_is_clear(UEINTX, NAKOUTI))
				;
			fifo_len = 0;
		}
	}
	if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
		goto ret;
	}
ret:
	if (fifo_len || (written != requested)) { // send ZLP
		bit_clear(UEINTX, TXINI);
	}
	// wait ZLP (from HOST)
	while (bit_is_clear(UEINTX, RXOUTI)
		&& bit_is_clear(UEINTX, NAKOUTI))
		;
	bit_clear(UEINTX, NAKOUTI);
	bit_clear(UEINTX, RXOUTI);
}

uint8_t reply_data[8];
uint8_t setup_data[8];
uint8_t usb_config_value;
uint8_t usb_send_empty;
uint8_t seq_no;


enum {
    SEND_ENCAPSULATED_COMMAND = 0,
    GET_ENCAPSULATED_RESPONSE,
    SET_COMM_FEATURE,
    GET_COMM_FEATURE,
    CLEAR_COMM_FEATURE,
    SET_LINE_CODING = 0x20,
    GET_LINE_CODING,
    SET_CONTROL_LINE_STATE,
    SEND_BREAK
};

uint8_t usbcdc_modebuf[7];
#define CDCMODE_BAUD0		0
#define CDCMODE_BAUD1		1
#define CDCMODE_BAUD2		2
#define CDCMODE_BAUD3		3
#define CDCMODE_STOPBIT		4
#define CDCMODE_PARITY		5
#define CDCMODE_DATABIT		6


int8_t usbcdc_setup(uint8_t *in, uint8_t *out) {
	if ( (in[USBRQ_TYPE] & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
	    if (in[USBRQ_REQ] == GET_LINE_CODING) {
		memcpy((void *)out, (const void *)usbcdc_modebuf, 7);
		return 7;
	    } else if (in[USBRQ_REQ] == SET_LINE_CODING) {
		int8_t i;
		while (bit_is_clear(UEINTX, RXOUTI))
			;
		for (i=0; i<7; i++) {
			usbcdc_modebuf[i] = UEDATX;
		}
		bit_clear(UEINTX, RXOUTI);
		return -1;  // send zlp
	    } else if (in[USBRQ_REQ] == SET_CONTROL_LINE_STATE) {
		usb_send_empty = 1;
		return -1;  // send zlp only
	    } else {
		return -2; // reply STALL
	    }
	} else {
		return -2; // reply STALL
	}
}


void usb_process_setup() {
	uint8_t request;
	uint8_t request_type;
	uint8_t rcpt;
	int8_t r;

	bit_clear(UEINTX,RXOUTI);
	for (r=0; r<8; r++) {
		setup_data[r] = UEDATX;
	}
	bit_clear(UEINTX,RXSTPI);

	request_type = setup_data[USBRQ_TYPE];
	request = setup_data[USBRQ_REQ];
	if (seq_no < 0x40) {
		uint8_t buf[4];
		buf[0] = request_type;
		buf[1] = request;
		buf[2] = setup_data[USBRQ_VALUE_L];
		buf[3] = setup_data[USBRQ_VALUE_H];
		eeprom_write_block(buf, (uint8_t *)seq_no, 4);
		seq_no +=4;
	}
	if (setup_data[USBRQ_LENGTH_H]) {
		reply_requested = 254;
	} else {
		reply_requested = setup_data[USBRQ_LENGTH_L];
	}
	if ((request_type & USBRQ_TYPE_MASK) == USBRQ_TYPE_STANDARD) {
		switch (request) {
		case USBRQ_GET_STATUS:
			rcpt = request_type & USBRQ_RCPT_MASK;
			reply_data[0] = 0;
			if (rcpt == USBRQ_RCPT_DEVICE) {
				reply_data[0] = USB_CFG_IS_SELF_POWERED;
			}
			if (rcpt == USBRQ_RCPT_ENDPOINT 
				&& setup_data[USBRQ_INDEX_L] == 0x81)
				reply_data[0] = 1;
			reply_data[1] = 0;
			usb_reply_setup(reply_data, 2, 0);
			break;
		case USBRQ_SET_ADDRESS:
#if 0
			if (!setup_data[USBRQ_VALUE_L] && !UDADDR) {
				bit_set(UECONX, STALLRQ);
				return;
			}
#endif
			UDADDR = setup_data[USBRQ_VALUE_L] & 0x7f;
			send_zlp();

			bit_set(UDADDR, ADDEN); // enable addrss
			break;
		case USBRQ_GET_DESCRIPTOR:
			if ( setup_data[USBRQ_VALUE_H] == 1) {
				usb_reply_setup((uint8_t *)usbDescrDevice,
						__LPM(usbDescrDevice), 1);
			} else if (setup_data[USBRQ_VALUE_H] == 2) {
				usb_reply_setup((uint8_t *)usbDescrConfig,
						__LPM(usbDescrConfig+2), 1);
			} else if (setup_data[USBRQ_VALUE_H] == 3) {
			    if (setup_data[USBRQ_VALUE_L] == 0) {
				usb_reply_setup((uint8_t *)usbDescrString0,
						__LPM(usbDescrString0), 1);
			    } else if (setup_data[USBRQ_VALUE_L] == 1) {
				usb_reply_setup((uint8_t *)usbDescrString1,
						__LPM(usbDescrString1), 1);
			    } else if (setup_data[USBRQ_VALUE_L] == 2) {
				usb_reply_setup((uint8_t *)usbDescrString2,
						__LPM(usbDescrString2), 1);
			    } else if (setup_data[USBRQ_VALUE_L] == 3) {
				usb_reply_setup((uint8_t *)usbDescrString3,
						__LPM(usbDescrString3), 1);
			    }
#if 0
			} else if (setup_data[USBRQ_VALUE_H] == USBDESCR_HID_REPORT) {
#endif
			} else {
				bit_set(UECONX, STALLRQ);
			}
			break;
		case USBRQ_GET_CONFIGURATION:
			reply_data[0] = usb_config_value;
			usb_reply_setup(reply_data, 1, 0);
			break;
		case USBRQ_SET_CONFIGURATION:
			usb_config_value = setup_data[USBRQ_VALUE_L];
			bit_clear(UEINTX, TXINI);  // send ZLP ?
			usb_connected = 1;
			break;
		case USBRQ_GET_INTERFACE:
			reply_data[0] = 0;
			usb_reply_setup(reply_data, 1, 0);
			break;
		case USBRQ_CLEAR_FEATURE:
		case USBRQ_SET_FEATURE:
			if (setup_data[USBRQ_VALUE_L] == 0 &&
				setup_data[USBRQ_INDEX_L] == 0x81) {
				/* EP1 HALT */
				bit_set(UECONX, STALLRQ);
			}
			bit_clear(UEINTX, TXINI); // send ZLP ?
			break;
		case USBRQ_SET_INTERFACE:
			bit_clear(UEINTX, TXINI); // send ZLP ?
			break;
		default:
			bit_set(UECONX, STALLRQ);
			break;
		}
	} else {
		r = usbcdc_setup(setup_data,reply_data);
		if (r >= 0) usb_reply_setup(reply_data, r, 0);
		else if (r == -1) send_zlp();
		else bit_set(UECONX, STALLRQ);
	}
}

static uint8_t log_seq;

void usbPoll() {
	uint8_t next_send_empty;
	UENUM = EP_CTRL;
	if (bit_is_set(UEINTX, RXSTPI)) { // recieve SETUP 
		usb_process_setup();
	}
	if (bit_is_set(UEINTX,NAKINI)) {
		bit_clear(UEINTX,NAKINI);
	}
	UENUM = EP_TX;
	if (bit_is_set(UEINTX,RXOUTI)) {
		bit_clear(UEINTX, RXOUTI);
	}
	if (bit_is_set(UEINTX, FIFOCON) && bit_is_clear(UEINTX,RWAL)) {
		bit_clear(UEINTX, FIFOCON);
	}
	UENUM = EP_RX;
	if (bit_is_set(UEINTX,TXINI)) {
		bit_clear(UEINTX,TXINI);
		bit_clear(UEINTX,NAKINI);
	}
	if (bit_is_set(UEINTX, FIFOCON)
		&& (usb_send_empty || bit_is_clear(UEINTX,RWAL) || 
		     (bit_is_set(UDINT, SOFI) && UEBCLX))) {
		if (log_seq < 0x40) {
			uint8_t data;
			data = (UESTA1X << 6) | (UEBCLX & 0x3f);
			eeprom_write_byte((uint8_t *)(0x40 + log_seq), data);
			log_seq += 1;
		}
		next_send_empty = UEBCLX;
		bit_clear(UEINTX, FIFOCON);
		usb_send_empty = next_send_empty;
	}
	if (bit_is_set(UDINT, SOFI)) {
		bit_clear(UDINT, SOFI);
	}
}

void usbcdc_init() {
	usbInit();
	sei();
	while (!usb_connected) {
		usbPoll();
	}
}

void usbcdc_poll() {
	usbPoll();
}

int8_t usbcdc_can_getc(int8_t size) {
	UENUM = EP_TX;
	return (UEBCLX >= size);
}

int8_t usbcdc_can_putc(int8_t size) {
	UENUM = EP_RX;
	return (UEBCLX + size) <= EP_RX_SIZE;
}

char usbcdc_getc(void) {
	char ret;
	UENUM = EP_TX;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbcdc_poll();
		UENUM = EP_TX;
		if (bit_is_set(UEINTX,RWAL) && (log_seq < 0x40)) {
			uint8_t data;
			data = (UESTA1X << 6) | 0x80 | (UEBCLX & 0x3f);
			eeprom_write_byte((uint8_t *)(0x40 + log_seq), data);
			log_seq += 1;
		}
	}
	ret = UEDATX;
	return ret;
}

void usbcdc_putc(char data) {
	UENUM = EP_RX;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbcdc_poll();
		UENUM = EP_RX;
	}
	UEDATX = data;
}
